unit xmlrpc_Browse;

{
  To do:
  - retrieve the type of each element
  - seperate Browser from Tester
  - editor: switch tabs: for navigating / use in text
            word-wrap yes / no
  - add code for actions in popup-menu
  - add nice graphics as used in browser interface
  - drag and drop to add files and images
  - view in browser (rendered by Zope!)
  - labels in tree should have id and title
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, Menus, ActnList, ExtCtrls, StdCtrls,
  // self
  xmlrpclib;

type
  TMainForm = class(TForm)
    PopupMenu1: TPopupMenu;
    ActionList1: TActionList;
    Edit: TAction;
    Edit_source: TMenuItem;
    Cut: TAction;
    Copy: TAction;
    Paste: TAction;
    Delete: TAction;
    Rename: TAction;
    Panel1: TPanel;
    Editor: TMemo;
    StatusBar1: TStatusBar;
    TreeView: TTreeView;
    Splitter1: TSplitter;
    Add_folder: TAction;
    Add_method: TAction;
    Add_document: TAction;
    Panel2: TPanel;
    Save: TButton;
    Panel3: TPanel;
    edtHost: TEdit;
    edtPort: TEdit;
    edtObject: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    edtUsername: TEdit;
    edtPassword: TEdit;
    Label4: TLabel;
    Label5: TLabel;
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure TreeViewDblClick(Sender: TObject);
    procedure EditExecute(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure SaveClick(Sender: TObject);
  private
    { Private declarations }
    Zope: TXMLRPC;
    sSend, sReceived: string;
    Params: TStringList;
    function GetFullPath( TreeNode: TTreeNode ): string;
    procedure GetObjectsRaw( TreeNode: TTreeNode );
    procedure GetContents( TreeNode: TTreeNode; sContents: TStrings );
    procedure PutContents( TreeNode: TTreeNode; sContents: string );
    procedure GetTitleAndId( TreeNode: TTreeNode; var sTitle, sId: string );
    function ParseRawToParams: Integer;
    function GetTag( var sString, sTag: string ): Integer;
    function GetData( var sString, sData: string ): Integer;
  public
    { Public declarations }
  end;
  function HTMLDecode( Raw: string ): string;
  function HTMLEncode( Clean: string ): string;

var
  MainForm: TMainForm;

implementation

uses
  StringLib;

{$R *.DFM}

procedure TMainForm.FormCreate(Sender: TObject);
begin
  Hide;
  Params := TStringList.Create;
  Zope := TXMLRPC.Create;
  Zope.Agent := 'Zope Stud 0.1';
  Zope.HostName := edtHost.Text;
  Zope.UserName := edtUsername.Text;
  Zope.Password := edtPassword.Text;
  Zope.Port := StrToInt( edtPort.Text );
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  Params.Free;
  Zope.Free;
end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  if TreeView.Selected = nil then begin
    TreeView.Items.Add( nil, edtObject.Text );
  end;
end;

procedure TMainForm.TreeViewDblClick(Sender: TObject);
// get the objects inside the current object
var
  TreeNode: TTreeNode;
  i: Integer;
begin
  // clear editor
  Editor.Lines.Clear;
  // do something
  TreeNode := TreeView.Selected;
  if not TreeNode.HasChildren then begin
    // get contents of the current node (first time only)
    GetObjectsRaw( TreeNode );
    if ParseRawToParams = 0 then begin
      // add the content as childs to current node
      for i := 0 to Params.Count-1 do
        TreeView.Items.AddChild( TreeNode, Params[ i ] );
      // expand current node
      TreeNode.Expanded := True;
    end
    else begin
      // MethodCall returned error; probably because it is not a folder; maybe Method or Document ?
      GetContents( TreeView.Selected, Editor.Lines );
    end;
  end;
end;

procedure TMainForm.EditExecute(Sender: TObject);
begin
  GetContents( TreeView.Selected, Editor.Lines );
end;

procedure TMainForm.SaveClick(Sender: TObject);
begin
  PutContents( TreeView.Selected, HTMLEncode( Editor.Lines.Text ) );
end;

function TMainForm.GetFullPath( TreeNode: TTreeNode ): string;
begin
  // retrieve the full path of the node
  Result := '';
  while TreeNode.Level > 0 do begin
    Result := TreeNode.Text + '/'+ Result;
    if TreeNode.Level > 0 then
      TreeNode := TreeNode.Parent;
  end;
end;

procedure TMainForm.GetObjectsRaw( TreeNode: TTreeNode );
begin
  sSend := '<?xml version="1.0"?>';
  sSend := sSend + '<methodCall>';
  sSend := sSend + '<methodName>objectIds</methodName>';
  sSend := sSend + '<params>';
  sSend := sSend + '</params>';
  sSend := sSend + '</methodCall>';
  Zope.Execute( GetFullPath( TreeNode ), sSend, sReceived );
end;

procedure TMainForm.GetContents( TreeNode: TTreeNode; sContents: TStrings );
begin
  sSend := '<?xml version="1.0"?>';
  sSend := sSend + '<methodCall>';
  sSend := sSend + '<methodName>document_src</methodName>';
  sSend := sSend + '<params>';
  sSend := sSend + '</params>';
  sSend := sSend + '</methodCall>';
  Zope.Execute( GetFullPath( TreeNode ), sSend, sReceived );
  if ParseRawToParams = 0 then
    sContents.Text := HTMLDecode( Params[ 0 ] );
end;

procedure TMainForm.GetTitleAndId( TreeNode: TTreeNode; var sTitle, sId: string );
var
  i, iLen: Word;
begin
  sSend := '<?xml version="1.0"?>';
  sSend := sSend + '<methodCall>';
  sSend := sSend + '<methodName>title_and_id</methodName>';
  sSend := sSend + '<params>';
  sSend := sSend + '</params>';
  sSend := sSend + '</methodCall>';
  Zope.Execute( GetFullPath( TreeNode ), sSend, sReceived );
  sTitle := ''; sId := '';
  if ParseRawToParams = 0 then begin
    // returns 'title (id)' or 'id'
    iLen := Length( Params[ 0 ] );
    if not ( Params[ 0 ][ iLen ] = ')' ) then
      sId := Params[ 0 ]
    else begin
      i := iLen;
      repeat
        i := i - 1;
      until Params[ 0 ][ i ] = '(';
      sTitle := Trim( System.Copy( Params[ 0 ], 1, i - 1 ) );
      sId := System.Copy( Params[ 0 ], i + 1, iLen - i - 1 );
    end;
  end
  else
    ; // MethodCall returned error
end;

procedure TMainForm.PutContents( TreeNode: TTreeNode; sContents: string );
// call: manage_edit(self, data, title)
// both params are needed
var
  sTitle, sId: string;
begin
  GetTitleAndId( TreeNode, sTitle, sId );
  sSend := '<?xml version="1.0"?>';
  sSend := sSend + '<methodCall>';
  sSend := sSend + '<methodName>manage_edit</methodName>';
  sSend := sSend + '<params>';
  sSend := sSend + '<param>';
  sSend := sSend + '<value><string>' + sContents + '</string></value>';
  sSend := sSend + '<value><string>' + sTitle + '</string></value>';
  sSend := sSend + '</param>';
  sSend := sSend + '</params>';
  sSend := sSend + '</methodCall>';
  Zope.Execute( GetFullPath( TreeNode ), sSend, sReceived );
  // returns the normal manage_edit browser screen as data (changes applied)
end;

function TMainForm.ParseRawToParams: Integer;
var
  sTag, sData: string;
begin
  Result := 0;
  Params.Clear;
  if ( Length( sReceived ) > 0 ) and not ( Uppercase( System.Copy( sReceived, 1, 6 ) ) = '<HTML>' ) then begin
    if ( Pos( '<name>faultCode</name>', sReceived ) > 0 ) then
      // MethodCall returned error
      Result := -1;
    while GetTag( sReceived, sTag ) = 0 do begin
      if Pos( sTag, '<i4><int><boolean><string><double><dateTime.iso8601><base64>' ) > 0 then begin
        // next comes data
        if GetData( sReceived, sData ) = 0 then
          Params.Append( sData )
        else begin
          Result := -2; // error parsing
          ShowMessage( 'func ParsRawToParams: error parsing'#10 + sReceived );
        end;
      end;
    end;
  end
  else
    // nothing returned or some html doc (browser error report)
    Result := -3;
end;

function TMainForm.GetData( var sString, sData: string ): Integer;
var
  iPos: Word;
begin
  Result := 0;
  if Length( sString ) > 0 then begin
    iPos := Pos( '<', sString );
    sData := System.Copy( sString, 1, iPos - 1 );
    System.Delete( sString, 1, iPos - 1 );
  end
  else
    // error: no more raw data
    Result := -1;
end;

function TMainForm.GetTag( var sString, sTag: string ): Integer;
var
  iPos: Word;
begin
  Result := 0;
  if sString[ 1 ] = #10 then
    System.Delete( sString, 1, 1 ); 
  if Length( sString ) > 0 then begin
    if sString[ 1 ] = '<' then begin
      iPos := Pos( '>', sString );
      sTag := System.Copy( sString, 1, iPos );
      System.Delete( sString, 1, iPos );
    end
    else
      // error parsing: tag expected but not found
      Result := -2;
  end
  else
    // error: no more raw data
    Result := -1;
end;

function HTMLEncode( Clean: string ): string;
// html needs to be encoded for transport
begin
  Result := strChange( strChange( Clean, '<', '&lt;' ), '<', '&gt;' );
end;

function HTMLDecode( Raw: string ): string;
// html was encoded for transport; now decode it again
begin
  Result := strChange( strChange( Raw, '&lt;', '<' ), '&gt;', '>' );
end;

end.
